
#define _GATAN_USE_STL_STRING		// Provide conversion from 'DM::String' to 'std::string'

#define _GATANPLUGIN_USES_LIBRARY_VERSION 2
#include "DMPlugInBasic.h"

#define _GATANPLUGIN_USE_CLASS_PLUGINMAIN
#include "DMPlugInMain.h"

using namespace Gatan;

#include <cassert>
#include <string>

class SampleImageProcessingPlugIn : public Gatan::PlugIn::PlugInMain
{
	virtual void Start();
	virtual void Run();
	virtual void Cleanup();
	virtual void End();
};

/*
** Implement the 'Horizontal_Derivative' script function.
*/
void ImgProc_Horizontal_Derivative( const DM::Image &src_img_in, DM::Image &l_deriv_img_out, DM::Image &r_deriv_img_out )
{
	/*
	** For ease, we will only deal with 32-bit floating data
	** ( 'ImageData::REAL4_DATA' in 'DMPlugInUtility.h' ), so
	** if the data is not already that type, create a temporary image
	** with the appropriate type.
	*/

	DM::Image src_img;
	if ( src_img_in.GetDataType() != ImageData::REAL4_DATA )
	{
		src_img = src_img_in.Clone();
		src_img.ChangeDataType( ImageData::REAL4_DATA );
	}

	/*
	** You can use the 'GetDimensionSize' methods of 'DM::Image' to 
	** get the sizes of dimensions instead of 'getsize'
	*/
	ulong xsize = src_img.GetDimensionSize( 0 );
	ulong ysize = src_img.GetDimensionSize( 1 );

	std::string name = src_img.GetName().get_string();

	// Create the destination images.  It looks like your script
	// did 'ConvertToShort' at the end, so we simply create the image
	// of type 'ImageData::SIGNED_INT16_DATA'.
	DM::Image l_deriv_img = DM::NewImage( ( std::string("Left Derivative of ") + name ).c_str()
	                                    , ImageData::SIGNED_INT16_DATA
										, xsize, ysize );

	DM::Image r_deriv_img = DM::NewImage( ( std::string("Right Derivative of ") + name ).c_str()
	                                    , ImageData::SIGNED_INT16_DATA
										, xsize, ysize );

	/*
	** Do the image processing
	** 
	*/

	{
		/*
		** The 'PlugIn::ImageDataLocker' class locks the data in the image down so
		** it is accessible from C++. The second parameter to the constructor specifies
		** how the data is to be locked down.  The data in the image might not be contiguous,
		** so the 'PlugIn::ImageDataLocker::lock_data_CONTIGUOUS' flag will force creation of
		** a contiguous temporary if necessary. If that flag is not specified, one must look
		** at the 'ImageData::image_data_t' structure returned by the method 'ImageDataLocker::get_image_data()'
		** to get the layout.  When the 'ImageDataLocker' object is destructed, the image data
		** is unlocked, and if 'MarkDataChanged' was called on the object, the data is copied back
		** to the source image if a temporary was created.
		** 
		*/
		PlugIn::ImageDataLocker src_img_l( src_img, PlugIn::ImageDataLocker::lock_data_CONTIGUOUS );
		PlugIn::ImageDataLocker l_deriv_img_l( l_deriv_img, PlugIn::ImageDataLocker::lock_data_CONTIGUOUS );
		PlugIn::ImageDataLocker r_deriv_img_l( r_deriv_img, PlugIn::ImageDataLocker::lock_data_CONTIGUOUS );

		/*
		** Here, we get the length and pointers to the data. The data is laid out in row-major order
		** ( dimension 0 first, then 1, 2, etc ), and the lengths and strides in each dimension
		** can be obtained by the 'image_data_t::get_length' and 'image_data_t::get_stride' methods
		** respectively.  The stride in a dimension is the number of bytes between the start of
		** consecutive data elements in that dimension. Because the data is locked so it is
		** contiguous, we may iterate through the data by ignore the stride, convert the pointer
		** to the data to a pointer to its C++ type, and index using the typed pointer.
		*/

		// Check that the data types are what we expect. They should be because we created them
		// with the desired type.
		assert( src_img_l.get_image_data().get_data_type() == ImageData::REAL4_DATA );
		assert( l_deriv_img_l.get_image_data().get_data_type() == ImageData::SIGNED_INT16_DATA );
		assert( r_deriv_img_l.get_image_data().get_data_type() == ImageData::SIGNED_INT16_DATA );

		// Get the lengths in the desired dimensions ( this is the same as 'xsize' and 'ysize' above ).
		ulong x_src_img_len = src_img_l.get_image_data().get_dim_length(0);
		ulong y_src_img_len = src_img_l.get_image_data().get_dim_length(1);

		// Get pointers to the data
		float32 *src_img_data     = reinterpret_cast<float32 *>( src_img_l.get_image_data().get_data() );
		sint16  *l_deriv_img_data = reinterpret_cast<sint16 *>( l_deriv_img_l.get_image_data().get_data() );
		sint16  *r_deriv_img_data = reinterpret_cast<sint16 *>( r_deriv_img_l.get_image_data().get_data() );

		/*
		** I am doing a simple, brain-dead iteration here. The compiler can probably optimize
		** this reasonably, but you could definitely hand-code this to be faster.  Also, the
		** rounding from float to int uses the c-style rounding ( round to zero ) and it doesn't
		** handle values which cannot fit in the integer properly, while DM uses round to nearest
		** and handles outliers properly.  We can get into how to do proper rounding later, but you
		** probably want to produce floating point final images anyway ( I am using the 'SIGNED_INT16_DATA'
		** type here to illustrate how to handle different data types ).
		*/
		for( ulong y_idx = 0; y_idx < y_src_img_len; ++y_idx )
		{
			// The end points are special cases, so do all but the end points here and
			for( ulong x_idx = 0; x_idx < ( x_src_img_len - 1 ); ++x_idx )
			{
				l_deriv_img_data[x_idx]   = (sint16) ( src_img_data[x_idx] - src_img_data[x_idx+1] );
				r_deriv_img_data[x_idx+1] = (sint16) ( src_img_data[x_idx+1] - src_img_data[x_idx] );
			}
			// Handle the endpoints
			l_deriv_img_data[x_src_img_len-1] = (sint16) src_img_data[x_src_img_len-1];
			r_deriv_img_data[0]               = (sint16) src_img_data[0];

			src_img_data     += x_src_img_len;
			l_deriv_img_data += x_src_img_len;
			r_deriv_img_data += x_src_img_len;
		}

		l_deriv_img_l.MarkDataChanged();
		r_deriv_img_l.MarkDataChanged();
	}

	l_deriv_img_out = l_deriv_img;
	r_deriv_img_out = r_deriv_img;
}

/*
** Provide a proxy for 'ImgProc_Horizontal_Derivative' that can be installed
** as a script function. Note that C++ class references cannot be passed to the
** script language, only raw pointers, so 'DM::Image' cannot be used as an argument
** to a script function. Instead, 'DM_ImageToken' is used for image rvalues, and
** 'DM_ImageToken *' for lvalues. The following function shows how to convert.
*/

void SF_ImgProc_Horizontal_Derivative( DM_ImageToken src_img, DM_ImageToken *l_deriv_img_out, DM_ImageToken *r_deriv_img_out )
{
	DM::Image l_deriv_img, r_deriv_img;
	ImgProc_Horizontal_Derivative( src_img, l_deriv_img, r_deriv_img );
	DM::Image::assign_ptr( *l_deriv_img_out, l_deriv_img.get() );
	DM::Image::assign_ptr( *r_deriv_img_out, r_deriv_img.get() );
}

///
/// This is called when the plugin is loaded.  Whenever DM is
/// launched, it calls 'Start' for each installed plug-in.
/// When it is called, there is no guarantee that any given
/// plugin has already been loaded, so the code should not
/// rely on scripts installed from other plugins.  The primary
/// use is to install script functions.
///
void SampleImageProcessingPlugIn::Start()
{
	AddFunction( "void ImgProc_Horizontal_Derivative( BasicImage src_img, BasicImage *l_deriv_img, BasicImage *r_deriv_img )", &SF_ImgProc_Horizontal_Derivative );
}

///
/// This is called when the plugin is loaded, after the 'Start' method.
/// Whenever DM is launched, it calls the 'Run' method for
/// each installed plugin after the 'Start' method has been called
/// for all such plugins and all script packages have been installed.
/// Thus it is ok to use script functions provided by other plugins.
///
void SampleImageProcessingPlugIn::Run()
{
}

///
/// This is called when the plugin is unloaded.  Whenever DM is
/// shut down, the 'Cleanup' method is called for all installed plugins
/// before script packages are uninstalled and before the 'End'
/// method is called for any plugin.  Thus, script functions provided
/// by other plugins are still available.  This method should release
/// resources allocated by 'Run'.
///
void SampleImageProcessingPlugIn::Cleanup()
{
}

///
/// This is called when the plugin is unloaded.  Whenever DM is shut
/// down, the 'End' method is called for all installed plugins after
/// all script packages have been unloaded, and other installed plugins
/// may have already been completely unloaded, so the code should not
/// rely on scripts installed from other plugins.  This method should
/// release resources allocated by 'Start', and in particular should
/// uninstall all installed script functions.
///
void SampleImageProcessingPlugIn::End()
{
}

SampleImageProcessingPlugIn gSampleImageProcessingPlugIn;